/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mitaka.h"
#include  "mtk_mode.h"
#include  "mtk_galaxies.h"
#include  "mtk_switches.h"
#include  "mtk_font.h"
#include  "mtk_OpenGL.h"
#include  "mtk_gl_common.h"
#include  "mtk_colors.h"
#include  "mtk_file.h"
#include  "mtk_solar_system.h"
#include  "mtk_elliptical_galaxy.h"
#include  "../common/astro/astrometry.h"
#include  <vector>
#include  <algorithm>



//
static int SDSS_Thinning = 0;

void
SetDistantGalaxyThinning(int val)
{
	if (val < 0)  val = 0;
	if (val > 10) val = 10;
	SDSS_Thinning = val;
}

struct distant_galaxy
{
	RA_DEC		eqPos;		// Right ascension/Declination (in degree)
	float		distance;	//  (in pc)
	vector3f	xyz;		// W
	float		ecls;		// e-class
};

struct QSO
{
	RA_DEC		eqPos;		// Right ascension/Declination (in degree)
	float		distance;	//  (in pc)
	vector3f	xyz;		// W
	float		z;			// red shift
};


vector<vector<distant_galaxy> >	FG_SDSS;	// SDSS Galaxies
vector<vector<QSO> >			vQSO;		// SDSS QSOs






//--------------------
//  Distant galaxies
//--------------------
bool
load_distant_galaxies(const directory& dir)
{
	const int class_N = 10;
	const float ecls_min = -0.35;
	const float ecls_max = 0.5;
	const float d_ecls = ecls_max - ecls_min;

	FG_SDSS.clear();
	FG_SDSS.resize( class_N );

	FILE* fp = dir.fopen("SDSS_GAL.dat", "rb");
	if (fp == NULL)  return false;

	int count = 0;
	while (true) {
		distant_galaxy gl;
		
		int n = fread(&gl.eqPos.RAdeg, sizeof(gl.eqPos.RAdeg),		1, fp);
		if (n != 1)  break;
		fread(&gl.eqPos.DEdeg,	sizeof(gl.eqPos.DEdeg),		1, fp);
		fread(&gl.distance,		sizeof(gl.distance),		1, fp);
		fread(&gl.ecls,			sizeof(gl.ecls),			1, fp);
		
		//-- ʒu߂
		vector3f Cxyz = gl.eqPos.get_dirc_vec();		// Vʏ̈ʒuiPʔaj
		gl.xyz  = M_R2E * ( gl.distance * Cxyz );		// SW

		//-- NX߂
		int icls = int( (gl.ecls - ecls_min) / d_ecls * class_N);
		if (icls < 0) icls = 0;
		if (icls >= class_N)  icls = class_N - 1;

		FG_SDSS[icls].push_back(gl);
		count++;
	}
	fclose(fp);

	return true;
}


//---  Position  ---
vector3d
get_distant_galaxy_position(int icls, int idx)
{
	return  FG_SDSS[icls][idx].xyz;
}



//--------
//  QSOs
//--------
bool
load_QSOs(const directory& dir)
{
	const int class_N = 20;
	const float z_min = 0.1;
	const float z_max = 1.75;
	const float dz = z_max - z_min;

	vQSO.clear();
	vQSO.resize( class_N );

	FILE* fp = dir.fopen("SDSS_QSO.dat", "rb");
	if (fp == NULL)  return false;

	int count = 0;
	while (true) {
		QSO q;
		
		int n = fread(&q.eqPos.RAdeg, sizeof(q.eqPos.RAdeg),		1, fp);
		if (n != 1)  break;
		fread(&q.eqPos.DEdeg,	sizeof(q.eqPos.DEdeg),	1, fp);
		fread(&q.distance,		sizeof(q.distance),		1, fp);
		fread(&q.z,				sizeof(q.z),			1, fp);
		
		//-- ʒu߂
		vector3f Cxyz = q.eqPos.get_dirc_vec();		// Vʏ̈ʒuiPʔaj
		q.xyz  = M_R2E * ( q.distance * Cxyz );		// SW

		//-- NX߂
		int icls = int( (q.z - z_min) / dz * class_N);
		if (icls < 0) icls = 0;
		if (icls >= class_N)  icls = class_N - 1;

		vQSO[icls].push_back(q);
		count++;
	}
	fclose(fp);

	return true;
}



void
draw_distant_galaxies(const mtk_draw_struct& ds, vector<vector<distant_galaxy> >& FGLX, float pt_sz)
{
	const float scale = ds.scale;
	if (ds.tyMode!=MODE_3D || scale < 1000)  return;
	const vector3d camera_pos = ds.camera.get_pos();

	//-- For culling
	GLdouble MM[16];
	glGetDoublev( GL_MODELVIEW_MATRIX,  MM);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);

	int skp = 1;
	float af = 1.0;
	pt_sz /= ds.rad_per_pixel / 0.001;
	if (pt_sz < 1) {
		af = pt_sz;
		pt_sz = 1;
	}
	glPointSize(pt_sz);

	// ߋ̃tF[h
	const float scl1 = 5.0e7;
	const float scl2 = 5.0e8;
	const float scl3 = 3.0e9;

	if (scale < scl1)  return;
	if (scale < scl2) {
		const float log_scl = log(scale);	
		const float log_scl1 = log(scl1);
		const float log_scl2 = log(scl2);
		af *= (log_scl - log_scl1) / (log_scl2-log_scl1);
	}

	// ̃tF[h
	if (scale > scl3) {
		float L = scale / scl3;

		af /= L*L;
	}
	af *= 0.7;
	if (af <= 0.00001)  return;
	while (af < 0.02) {
		af *= 2;
		skp *= 2;
	}

	// Thinning out
	for (int i=0; i<SDSS_Thinning; i++) {
		af *= 2;
		skp *= 2;
	}

	if (af > 1) {
		af = 1;
	}

	glPushMatrix();
	glTranslated(-camera_pos.x, -camera_pos.y, -camera_pos.z);






	const int class_N = FGLX.size();
	for (int ic=0; ic<class_N; ic++) {
		float t = ic/float(class_N-1);
		float B_V = 2.2*(1-t) - 0.2;
		RGBf col = correct_RGB( 0.9f * blackbody_color( B_V_to_Tc( B_V ) ) );
		glColor4f( col.r, col.g, col.b, af );
		const vector<distant_galaxy>&  GAL = FGLX[ic];
		


		if (false) {
			// QUAD
			//---  JWn̎
			vector3d  eF, eU, eR;
			ds.camera.get_axes(eF, eU, eR);
			const vector3d rpos = -ds.camera.get_pos();
			const double a = 0.3 / get_screen_z() * get_screen_w() / 180;

			for (int i=0; i<GAL.size(); i+=skp) {
				const distant_galaxy& gl = GAL[i];
				vector3f  rr = gl.xyz + rpos;

				//---  lpWn̎
				double R2;
				R2 = norm2(rr);
				if (R2 < 1e-100)  continue;
				
				vector3d  eF2, eU2, eR2;
				double R;
				R = sqrt(R2);
				eF2 = rr / R;
				eR2 = normalize( cross_product(eF2, eU) );
				eU2 = cross_product( eR2, eF2 );

				double sz;
				sz = a * R;

				vector3d  r = rr - 0.5*sz*eR2 - 0.5*sz*eU2;
				glBegin(GL_QUADS);
					glVertex3dv( r.get_pointer() );
					r += sz * eR2;
					glVertex3dv( r.get_pointer() );
					r += sz * eU2;
					glVertex3dv( r.get_pointer() );
					r -= sz * eR2;
					glVertex3dv( r.get_pointer() );
				glEnd();
			}
		} else {
			glBegin(GL_POINTS);
			for (int i=0; i<GAL.size(); i+=skp) {
				const distant_galaxy& gl = GAL[i];
				glVertex3fv(gl.xyz.get_pointer());
			}
			glEnd();
		}
	}
	glPopMatrix();


	glEnable(GL_DEPTH_TEST);
	glPointSize(1);
}


void
draw_QSOs(const mtk_draw_struct& ds, float pt_sz)
{
	const float scale = ds.scale;
	if (ds.tyMode!=MODE_3D || scale < 1000)  return;
	const vector3d camera_pos = ds.camera.get_pos();

	//-- For culling
	GLdouble MM[16];
	glGetDoublev( GL_MODELVIEW_MATRIX,  MM);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);

	int skp = 1;
	float af = 1.0;
	pt_sz /= ds.rad_per_pixel / 0.001;
	if (pt_sz < 1) {
		af = pt_sz;
		pt_sz = 1;
	}
	glPointSize(pt_sz);

	// ߋ̃tF[h
	const float scl1 = 3.0e8;
	const float scl2 = 1.9e10;
	if (scale < scl1)  return;
	if (scale < scl2) {
		const float log_scl = log(scale);	
		const float log_scl1 = log(scl1);
		const float log_scl2 = log(scl2);
		af *= (log_scl - log_scl1) / (log_scl2-log_scl1);
	}

	// ̃tF[h
	if (scale > scl2) {
		float L = scale / scl2;
		af /= L*L;
	}
	af *= 0.7;
	if (af <= 0.00001)  return;
	while (af < 0.02) {
		af *= 2;
		skp *= 2;
	}

	// Thinning out
	for (int i=0; i<SDSS_Thinning; i++) {
		af *= 2;
		skp *= 2;
	}

	if (af > 1) {
		af = 1;
	}

	glPushMatrix();
	glTranslated(-camera_pos.x, -camera_pos.y, -camera_pos.z);
	const RGBf col1 = get_color( MTK_COLOR_QSO );
	const RGBf col2 = get_color( MTK_COLOR_QSO );
	const int class_N = vQSO.size();
	const float z_min = 0.1;
	const float z_max = 1.75;
	for (int ic=0; ic<class_N; ic++) {
		float t = ic/float(class_N-1);
		RGBf col = (1-t)*col1 + t*col2;
		glColor4f( col.r, col.g, col.b, af );
		const vector<QSO>&  Q = vQSO[ic];
		
		glBegin(GL_POINTS);
		for (int i=0; i<Q.size(); i+=skp) {
			const QSO& q = Q[i];
			glVertex3fv(q.xyz.get_pointer());
		}
		glEnd();
	}
	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	glPointSize(1);
}



void
draw_galaxies(const mtk_draw_struct& ds)
{
	if (swGalaxy & GLXY_DISTANT_GALAXY)		draw_distant_galaxies(ds, FG_SDSS, 1.7);
}

void
draw_QSOs(const mtk_draw_struct& ds)
{
	draw_QSOs(ds, 1.7);
}


//----------
//  f[^
//----------
int
get_number_of_distant_galaxies()
{
	int count = 0;
	for (int i=0; i<FG_SDSS.size(); i++) {
		count += FG_SDSS[i].size();
	}

	return  count;
}

int
get_number_of_QSOs()
{
	int count = 0;
	for (int i=0; i<vQSO.size(); i++) {
		count += vQSO[i].size();
	}

	return  count;
}
