/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "../common/math/vector3.h"
#include  "../common/math/quaternion.h"
#include  "../common/math/spline.h"
#include  "../common/graphics/image.h"
#include  "../common/graphics/dib.h"
#include  "../common/graphics/avi.h"
#include  "../common/graphics/gl_win.h"
#include  "../common/graphics/frame.h"
#include  "mitaka.h"
#include  "mtk_make_movie.h"
#include  "mtk_switches.h"
#include  "mtk_text.h"
#include  "mtk_time.h"
#include  "mtk_target.h"
#include  "mtk_camera_setting.h"
#include  "mtk_stars.h"
#include  "mtk_draw_struct.h"
#include  "mtk_screen.h"
#include  "mtk_gl_common.h"
#include  "mtk_cubic_image.h"


//-----------
//  Externs
//-----------
extern HWND		g_hWnd;
extern COpenGL	ogl;
extern int win_w, win_h;
extern int  gl_w, gl_h;
extern mtk_camera g_camera;
extern float FontSize;
extern int animation_counter0;
extern int animation_counter;
extern int animation_interval;
extern int animation_type;

//-----------------------
//  Defined in main.cpp
//-----------------------
void		draw_scene(mtk_draw_struct ds);
void		OnSetZoom(double val);
void		OnCommandDraw();
void		gl_capture_viewport(image& img);

//--- Path ---
int path_frames=0;
int current_frame=0;
vector<path_info>  pathif(1000);
int path_interval = 100;
int g_path_type = 0;
int g_path_fix = 0;

//--- Make Movie ---
bool  bPlayFrame = false;
int   play_frame;
int   play_frame_sub;
bool  bRecordMovie = false;
int   record_frame;
int   record_count;
int   avi_frame;

vector3d
calc_camera_pos(const vector3d& target_pos, const quaterniond& ori, const double scale)
{
	mtk_camera  cam;
	cam.set_ori(ori);
	cam.set_scale_factor( calc_scale_factor(scale) );
	const double dist = get_obj_z() / cam.get_scale_factor();
	cam.set_pos_by_target_pos(target_pos, dist);
	return cam.get_pos();
}

//---------
//  
//---------
void
init_path()
{
	for (int i=0; i<pathif.size(); ++i) {
		pathif[i].interval = 50;
	}
}


//------------------
//  t[̃Zbg
//------------------
void init_interpolation();


void
set_frame(const path_info& pif)
{
}


//-------------------
//  t[Ԃ̏
//-------------------
// 
vector<double>	spl_t;
// pX̋
vector<double>	spl_l, spl_dl0, spl_dl;		// linear
vector<double>	spl_s, spl_ds0, spl_ds;		// log
// ^[Qbgʒu
vector<double>	spl_tx,  spl_ty,  spl_tz;
vector<double>	spl_tx2, spl_ty2, spl_tz2;
// Jʒu
vector<double>	spl_cx,  spl_cy,  spl_cz;
vector<double>	spl_cx2, spl_cy2, spl_cz2;
// 邳
vector<double>	spl_logLref, spl_logLref2;


void
init_interpolation()
{
	const int n = path_frames;

	// resize
	if (spl_t.size() != n) {
		spl_t.resize(n);
		spl_l.resize(n);
		spl_dl0.resize(n);
		spl_dl.resize(n);
		spl_s.resize(n);
		spl_ds0.resize(n);
		spl_ds.resize(n);
		spl_logLref.resize(n);
		spl_logLref2.resize(n);
	}

	//-- time
	int   it=0;
	float t=0.0;
	for (int i=0; i<n; ++i) {
		pathif[i].time = spl_t[i] = float(it);
		it += pathif[i].interval;
	}

	//-- camera position
	for (int i=0; i<n; ++i)		{
		path_info& pif = pathif[i];
		pif.camera_pos = calc_camera_pos(pif.target_pos, pif.ori, pif.scale);
	}

	//-- path length
	float l = 0.0;
	spl_l[0] = 0.0;
	spl_s[0] = 0.0;
	for (int i=0; i<n-1; ++i) {
		path_info& pif1 = pathif[i];
		path_info& pif2 = pathif[i+1];
		float dl = norm(pif2.camera_pos - pif1.camera_pos);
		l += dl;
		spl_l[i+1] = l;
		spl_s[i+1] = log(1.0 + l);
	}

	//-- path length ̔vZ
	for (int i=0; i<n-1; i++) {
		float dt = spl_t[i+1] - spl_t[i];
		spl_dl0[i] = (spl_l[i+1] - spl_l[i]) / dt;
		spl_ds0[i] = (spl_s[i+1] - spl_s[i]) / dt;
	}

	//-- ZOgEł̔vZ
	for (int i=1; i<n-1; i++) {
		float w1 = spl_l[i]   - spl_l[i-1];
		float w2 = spl_l[i+1] - spl_l[i];
		float wt = w1 + w2;
		spl_dl[i] = (w2/wt *spl_dl0[i-1] + w1/wt * spl_dl0[i]);
		w1 = spl_s[i]   - spl_s[i-1];
		w2 = spl_s[i+1] - spl_s[i];
		wt = w1 + w2;
		spl_ds[i] = (w2/wt *spl_ds0[i-1] + w1/wt * spl_ds0[i]);
	}

	// [_͎RE
	spl_dl[0]   = spl_dl0[0];
	spl_dl[n-1] = spl_dl0[n-2];
	spl_ds[0] = spl_ds0[0];
	spl_ds[n-1] = spl_ds0[n-2];

	//-- L_ref2
	for (int i=0; i<n; ++i)		{ spl_logLref[i] = log10(pathif[i].L_ref); }
	spline(spl_t,spl_logLref,spl_logLref2);
}



//----------------
//  t[̕
//----------------
path_info
intr_frame(const double t)
{
	path_info pif;

	// piʐ`ԁj
	int i;
	for (i=0; i<path_frames; ++i) {
		if (t < pathif[i].time)  break;
	}
	const path_info& pif1 = pathif[i-1];
	const path_info& pif2 = pathif[i];

	// Ԃ̎
	double t1 = spl_t[i-1];
	double t2 = spl_t[i];
	double dt = t2 - t1;
	double te = t - t1;
	double te2 = te*te;
	double te3 = te2*te;
	double tt;

	if (pif1.ty == 2) {					//--- log_cubic
		float s1  = spl_s[i-1];
		float s2  = spl_s[i];
		float ds1 = spl_ds[i-1];
		float ds2 = spl_ds[i];
		float ds = s2 - s1;
		float a = (ds2 + ds1 - 2.0*ds/dt) / (dt*dt);
		float b = (3.0*ds/dt - ds2 - 2.0*ds1)/dt;
		float c = ds1;
		tt = a * te3 + b * te2 + c * te;
		tt /= ds;
	} else if (pif1.ty == 1) {				//--- cubic
		float l1  = spl_l[i-1];
		float l2  = spl_l[i];
		float dl1 = spl_dl[i-1];
		float dl2 = spl_dl[i];
		float dl = l2 - l1;
		float a = (dl2 + dl1 - 2.0*dl/dt) / (dt*dt);
		float b = (3.0*dl/dt - dl2 - 2.0*dl1)/dt;
		float c = dl1;
		tt = a * te3 + b * te2 + c * te;
		tt /= dl;
	} else {	//--- linear
		tt = te / dt;
	}

	if (tt < 0.0) tt = 0.0;
	if (tt > 1.0) tt = 1.0;
	pif.ori	= slerp( pif1.ori, pif2.ori, tt );
	
	// XP[
	double log_scl1 = log10(pif1.scale);
	double log_scl2 = log10(pif2.scale);
	double log_scale = (1-tt) * log_scl1 + tt * log_scl2;
	pif.scale = pow(10.0, log_scale);

	// RXvC
	pif.L_ref = pow(10.0, splint(spl_t, spl_logLref, spl_logLref2, t));

	// ^[QbgƃJ̈ʒu
	vector3f  target_pos, camera_pos, camera_pos0;

	switch (pif1.fix) {
		case 1:	// JŒ
			camera_pos0 = calc_camera_pos(pif1.target_pos, pif1.ori, pif1.scale);
			target_pos	= (1-tt) * pif1.target_pos + tt * pif2.target_pos;
			camera_pos  = calc_camera_pos(target_pos, pif.ori, pif.scale);
			target_pos += camera_pos0 - camera_pos;
			camera_pos  = camera_pos0;

			break;
		case 0:
		default:
			target_pos	= (1-tt) * pif1.target_pos + tt * pif2.target_pos;
			camera_pos = calc_camera_pos(target_pos, pif.ori, pif.scale);
			break;
	}

	pif.target_pos = target_pos;
	pif.camera_pos = camera_pos;

	return pif;
}



//---------------
//  Play frames
//---------------
void
draw_intr_frame(const float t)
{
	/****** II
	//-- Clear
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	path_info  pif = intr_frame(t);
	set_frame(pif);
	vector3d tpos = pif.target_pos;

	// camera
	mtk_camera cam;
	cam.set_ori(pif.ori);
	set_target_pos(tpos);
	cam.set_scale_factor( calc_scale_factor( pif.scale ) );
	const double dist = ini_obj_z / cam.get_scale_factor();
	cam.set_pos_by_target_pos(tpos, dist);

	// draw struct
	mtk_draw_struct ds;
	ds.set(cam, get_time_JD(), gl_w, gl_h, FontSize);
	draw_scene(ds);
	*/
}

void
PlayFrame()
{
	float t = pathif[play_frame].time + float(play_frame_sub);
	draw_intr_frame(t);
}


//--------
//  Path
//--------
void
OnLoadPath()
{
	FILE* fp;
	if ((fp=fopen("path.dat", "rt")) == NULL)  return;
	double t;
	fscanf(fp, "JD = %lf\n\n", &t);
	change_time_JD(t);

	int count = 0;
	while (true) {
		path_info& pif = pathif[count];
		int n, dm;

		//-- frames
		n = fscanf(fp, "Frame = %d\n", &dm);
		if (n==EOF || n==0) break;
		//-- Target position
		fscanf(fp, "tp:%le %le %le\n",
					&pif.target_pos.x, &pif.target_pos.y, &pif.target_pos.z);
		//-- Camera position
		fscanf(fp, "cm:%le %le %le\n",
					&pif.camera_pos.x, &pif.camera_pos.y, &pif.camera_pos.z);
		//-- Orientation
		quaterniond& q = pif.ori;
		fscanf(fp, "or:%le %le %le %le\n", &q.w, &q.v.x, &q.v.y, &q.v.z);
		//-- Scale
		fscanf(fp, "sc:%e\n", &pif.scale);
		//-- L_ref
		fscanf(fp, "Lr:%e\n", &pif.L_ref);
		//-- interval
		fscanf(fp, "it:%d\n", &pif.interval);
		//-- type
		fscanf(fp, "ty:%d\n", &pif.ty);
		//-- fix
		fscanf(fp, "fx:%d\n\n", &pif.fix);

		count++;
	}
	path_frames = count;

	OnCommandDraw();
}


void
OnSavePath()
{
	if (path_frames == 0) return;

	FILE* fp;
	if ((fp=fopen("path.dat", "wt")) == NULL)  return;

	fprintf(fp, "JD = %lf\n\n", get_time_JD());

	for (int i=0; i<path_frames; ++i) {
		path_info& pi = pathif[i];

		//-- frames
		fprintf(fp, "Frame = %d\n", i);

		//-- Target position
		fprintf(fp, "tp:%le %le %le\n", pi.target_pos.x, pi.target_pos.y, pi.target_pos.z);

		//-- Camera position
		fprintf(fp, "cm:%le %le %le\n", pi.camera_pos.x, pi.camera_pos.y, pi.camera_pos.z);

		//-- Orientation
		const quaterniond& q = pi.ori;
		fprintf(fp, "or:%le %le %le %le\n", q.w, q.v.x, q.v.y, q.v.z);

		//-- Scale
		fprintf(fp, "sc:%le\n", pi.scale);

		//-- L_ref
		fprintf(fp, "Lr:%le\n", pi.L_ref);

		//-- interval
		fprintf(fp, "it:%d\n", pi.interval);

		//-- type
		fprintf(fp, "ty:%d\n", pi.ty);

		//-- fix
		fprintf(fp, "fx:%d\n\n", pi.fix);
	}

	fclose(fp);
}

void
OnClearPath()
{
	path_frames = 0;
	current_frame = 0;
	OnCommandDraw();
}


void
OnChangeCurrentFrame(int frm)
{
	if (!bPathInfo)  return;
	if (frm < 0 || frm > path_frames)  return;
	current_frame = frm;
	path_interval = pathif[current_frame].interval;
	OnCommandDraw();
}

void
OnChangeFrameInterval(int d)
{
	if (!bPathInfo)  return;
	int itvl = path_interval;
	itvl += d;
	if (itvl < 1) return;
	path_interval = itvl;
	OnCommandDraw();
}

void
OnChangePathType()
{
	if (!bPathInfo)  return;
	g_path_type++;
	if (g_path_type >= 3) g_path_type = 0;
	OnCommandDraw();
}


void
OnLoadFrame()
{
	if (!bPathInfo)  return;
	set_frame(pathif[current_frame]);
	OnCommandDraw();
}

void
OnSetFrame()
{
	/***** !!
	if (!bPathInfo)  return;
	path_info&  pifc = pathif[current_frame];
	pifc.target_pos	= get_current_target_pos();
	pifc.ori		= g_camera.get_ori();
	pifc.scale		= get_current_scale();
	pifc.L_ref		= get_L_ref();
	pifc.interval	= path_interval;
	pifc.ty			= g_path_type;
	pifc.camera_pos	= g_camera.get_pos();
	pifc.fix		= g_path_fix;

	if (current_frame==path_frames) {
		path_frames++;
		current_frame = path_frames;
		pathif[current_frame] = pathif[current_frame-1];
		path_interval = 50;
	} else {
		OnChangeCurrentFrame(current_frame+1);
		OnLoadFrame();
		return;
	}
	OnCommandDraw();
	*/
}

void
OnInsertFrame()
{
	/****  II
	if (!bPathInfo)  return;

	path_info pif;
    pif.target_pos	= get_current_target_pos();
	pif.ori			= g_camera.get_ori();
	pif.scale		= get_current_scale();
	pif.L_ref		= get_L_ref();
	pif.interval	= path_interval;
	pif.ty			= g_path_type;
	pif.fix			= g_path_fix;

	vector<path_info>::iterator  it = pathif.begin() + current_frame;
	pathif.insert(it, pif);
	path_frames++;
	OnCommandDraw();
	*/
}

void
OnDeleteFrame()
{
	if (!bPathInfo)  return;

	vector<path_info>::iterator  it = pathif.begin() + current_frame;
	pathif.erase(it);
	path_frames--;
	OnCommandDraw();
}

//--------------
//  Play Frame
//--------------
bool
start_play_frame(bool bFlg)
{
	if (path_frames==0)  return false;

	bPlayFrame = bFlg;
	play_frame = 0;
	animation_counter = pathif[play_frame].interval - 1;
	animation_interval = 1;
	animation_type = ANIM_TY_PLAY;
	init_interpolation();

	return true;
}


//--------------
//  Make Movie
//--------------
AVIFile  avi;
image  mvimg;
dib    mvdib;

bool
openAVI(const char* filename)
{
	dib ofs(win_w, win_h, 24);


	UINT rate = 30;
	AVIFileInit();
	avi.open(filename, OF_CREATE | OF_WRITE | OF_SHARE_DENY_WRITE );
	avi.createVIDEO( win_w, win_h, 1, rate);
	avi.selectCompression();
	avi.setFormat(ofs.get_pbmih());

	return  true;
}

void
closeAVI()
{
	avi.close();
	AVIFileExit();
}

void
InitRecordingMovie()
{
	bRecordMovie = false;
	
	if (make_movie_type == MV_AVI) {
		if (!openAVI(mv_filename)) { closeAVI(); return; }
	}

	mvimg.alloc(win_w, win_h, 3);
	mvdib.alloc(win_w, win_h, 24);

	avi_frame = 0;
	record_frame = 0;
	record_count = 0;
	bRecordMovie = true;
}

void
RecordMovie()
{
	const int R = 1;
	const path_info& pif1 = pathif[record_frame];
	const path_info& pif2 = pathif[record_frame+1];
	float t = pathif[record_frame].time + float(record_count)/R;

	if (make_movie_type == MV_AVI) {
		//------------- AVI t@C
		HDC hDC = GetDC(g_hWnd);
		
		ogl.MakeCurrent(hDC);
			draw_intr_frame(t);
			SwapBuffers(hDC);
		
			gl_capture_viewport(mvimg);
			image_to_dib(mvimg, mvdib);
			avi.writeFrame(avi_frame, mvdib.get_pBits());
			avi_frame++;
		ogl.MakeCurrent(NULL);

		ReleaseDC(g_hWnd, hDC);
	} else if (make_movie_type == MV_TARGA || make_movie_type == MV_PNG) {
		//----------- TARGA Aԃt@C
		char filename[256];
		if (make_movie_type == MV_TARGA) {
			sprintf(filename, "%s%05d.tga", mv_basefilename, avi_frame);
		} else {
			sprintf(filename, "%s%05d.png", mv_basefilename, avi_frame);
		}

		HDC hDC = GetDC(g_hWnd);
		ogl.MakeCurrent( hDC );
			draw_intr_frame(t);
			SwapBuffers(hDC);
			
			gl_capture_viewport(mvimg);
			if (make_movie_type == MV_TARGA)  { mvimg.saveTGA(filename); }
			else							  { mvimg.savePNG(filename); }
			avi_frame++;
		ogl.MakeCurrent(NULL);

		ReleaseDC(g_hWnd, hDC);
	} else if (make_movie_type == MV_CUBIC) {
		quaterniond	tmp_ori		= get_local_camera_orientation();
		float tmp_eye_offset_x	= get_eye_offset_x();
		HDC hDC = GetDC(g_hWnd);
		ogl.MakeCurrent( hDC );
			make_cubic_frame(mv_basefilename, t, avi_frame, 1024, hDC);
		ogl.MakeCurrent(NULL);

		avi_frame++;
		set_local_camera(tmp_ori, vector3d(0,0,0));
		set_eye_offset_x( tmp_eye_offset_x );
	} else {
		//--------- TARGA AԂUʓ
		quaterniond	tmp_ori		= get_local_camera_orientation();
		float tmp_eye_offset_x	= get_eye_offset_x();
		float eye_angle = -45;
		char scr_clr[] = { 'L', 'C', 'R' };
		char eye_lr[] = { 'L', 'R' };
		float eye_offset_x_abs = fabs(tmp_eye_offset_x);
		

		for (int i=0; i<3; i++) {
			set_local_camera( 45*(i-1), 0 );
			for (int j=0; j<2; j++) {
				char filename[256];
				if (make_movie_type == MV_TARGA6) {
					sprintf(filename, "%s_%c%c_%05d.tga", mv_basefilename, scr_clr[i], eye_lr[j], avi_frame);
				} else {
					sprintf(filename, "%s_%c%c_%05d.png", mv_basefilename, scr_clr[i], eye_lr[j], avi_frame);
				}

				set_eye_offset_x((2*j-1)*eye_offset_x_abs);
				HDC hDC = GetDC(g_hWnd);
				ogl.MakeCurrent( hDC );
					draw_intr_frame(t);
					SwapBuffers(hDC);
					gl_capture_viewport(mvimg);
					if (make_movie_type == MV_TARGA6)  { mvimg.saveTGA(filename); }
					else							   { mvimg.savePNG(filename); }
				ogl.MakeCurrent(NULL);
				ReleaseDC(g_hWnd, hDC);
			}
		}
		avi_frame++;

		set_local_camera(tmp_ori, vector3d(0,0,0));//*****
		set_eye_offset_x( tmp_eye_offset_x );
	}

	// To next frame
	record_count++;
	if (record_count >= pif1.interval*R) {
		record_frame++;
		record_count=0;
	}
	if (record_frame >= path_frames-1) {
		closeAVI();
		mvimg.free();
		mvdib.free();
		bRecordMovie = false;
	}
}


//---------------------
//  pX̋O\
//---------------------
void
draw_path()
{
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

    // pX̕\
	if (path_frames >= 2) {
		init_interpolation();

		//-- ^[Qbg̈ʒu
		glLineWidth(1.0);
		glColor3f(1,1,0);
		glBegin(GL_LINE_STRIP);
		for (int i=0; i<path_frames-1; ++i) {
			float t = pathif[i].time;
			for (int j=0; j<pathif[i].interval; ++j) {
			    path_info pif = intr_frame(t);
				t += 1.0;
				vector3d  pos = pif.target_pos; 
				glVertex3dv(pos.get_pointer());
			}
		}
		glEnd();

		//-- J̈ʒu
		glLineWidth(1.0);
		glColor3f(0,1,0);
		glBegin(GL_LINE_STRIP);
		for (int i=0; i<path_frames-1; ++i) {
			float t = pathif[i].time;
			for (int j=0; j<pathif[i].interval; ++j) {
			    path_info pif = intr_frame(t);
				t += 1.0;
				vector3d  pos = pif.camera_pos;
				glVertex3dv(pos.get_pointer());
			}
		}
		glEnd();

		glPointSize(5);
		glColor3f(1,1,0);
		glBegin(GL_POINTS);
		for (int i=0; i<path_frames-1; ++i) {
			float t = pathif[i].time;
			for (int j=0; j<pathif[i].interval; ++j) {
			    path_info pif = intr_frame(t);
				t += 1.0;
				vector3d  pos = pif.target_pos;
				glVertex3dv(pos.get_pointer());
			}
		}
		glEnd();


		for (int i=0; i<path_frames-1; ++i) {
			for (int j=0; j<pathif[i].interval; j+=2) {
				float t = pathif[i].time + float(j);
			    path_info pif = intr_frame(t);

				vector3d  pos = pif.camera_pos;			
				if (j==0)  {
					glPointSize(7);
                    glColor3f(1,0,0);
				} else {
					glPointSize(5);
                    glColor3f(0,1,0);
				}
				glBegin(GL_POINTS);
					glVertex3dv(pos.get_pointer());
				glEnd();
			}
		}

		// target
		glPointSize(5);
		glColor3f(1,0,0);
		glBegin(GL_POINTS);
			glVertex3f(0,0,0);		
		glEnd();
	}
}

//-----------------------
//  pX\
//-----------------------
void
draw_path_info()
{
	/******** II
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	const float alpha = 0.6f;
	glPushMatrix();
		goto_center_screen();
		const float scl0 = get_screen_w() / 180.0;
		glScalef(scl0,scl0,scl0);
		glColor4f(0.4,0.9,0.4,alpha);
		float sz = 6;
		glScalef(sz,sz,sz);

		char str[256];

		glTranslatef(0,-4,0);
		if (path_frames==0) {
			sprintf(str, "No Frames");
			draw_text_line(str, TEXT_CENTER, false);
			glPopMatrix();
			glEnable(GL_DEPTH_TEST);
			return;
		}

		float scl, L_ref;
		int itvl, ty, fix;
		vector3d pos;
		quaterniond ori;
		if (current_frame == path_frames || bPlayFrame) {
			pos = get_current_target_pos();
			ori = g_camera.get_ori();
			scl = get_current_scale();
			L_ref = get_L_ref();
			itvl = path_interval;
			ty = g_path_type;
			fix = g_path_fix;
		} else {
            path_info& pi = pathif[current_frame];
			pos = pi.target_pos;
			ori = pi.ori;
			scl = pi.scale;
			L_ref = pi.L_ref;
			itvl = pi.interval;
			ty = pi.ty;
			fix = pi.fix;
		}

		if (current_frame==path_frames) {
			glColor4f(0.4,0.6,0.9,alpha);
		}
		sprintf(str, "Current frame = %d   Frames = %d", current_frame, path_frames);
		draw_text_line(str, TEXT_CENTER, true);

		glColor4f(0.4,0.9,0.4,alpha);
		sprintf(str, "Target Pos = %.2le %.2le %.2le", pos.x, pos.y, pos.z);
		draw_text_line(str, TEXT_CENTER, true);

		sprintf(str, "Orientation = %.2le %.2le %.2le %.2le",
				ori.w, ori.v.x, ori.v.y, ori.v.z);
		draw_text_line(str, TEXT_CENTER, true);

		sprintf(str, "Scale = %le pc   L_ref = %le", scl, L_ref);
		draw_text_line(str, TEXT_CENTER, true);

		sprintf(str, "Interval to next frame = %d [%d]", itvl, path_interval);
		draw_text_line(str, TEXT_CENTER, true);

		// ^Cv
		const char szType[3][10] = {"Linear", "Cubic", "Log Cubic" };
		sprintf(str, "Type = %s", szType[ty]);
		draw_text_line(str, TEXT_CENTER, true);

		// Œ肷邩
		const char szFix[2][10] = {"Target", "Camera" };
		sprintf(str, "Fix = %s", szFix[fix]);
		draw_text_line(str, TEXT_CENTER, true);

		if (bPlayFrame) {
			sprintf(str, "Play Frame = %d - %d/%d", play_frame, play_frame_sub, pathif[play_frame].interval);
		draw_text_line(str, TEXT_CENTER, false);
		}
	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	*/
}

