/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_object.h"
#include  "mtk_mode.h"
#include  "mtk_object_position.h"
#include  "mtk_solar_system.h"
#include  "mtk_sun.h"
#include  "mtk_planets.h"
#include  "mtk_satellites.h"
#include  "mtk_satellite_data.h"
#include  "mtk_planet_position.h"
#include  "mtk_planet_orientation.h"
#include  "mtk_planet_phys_params.h"
#include  "mtk_minor_planets.h"
#include  "mtk_spacecraft.h"
#include  "mtk_oort_cloud.h"
#include  "mtk_fixed_object.h"
#include  "mtk_stars.h"
#include  "mtk_stars_2.h"
#include  "mtk_globular_clusters.h"
#include  "mtk_our_galaxy.h"
#include  "mtk_nearby_galaxies.h"
#include  "mtk_galaxies.h"	//**** old
#include  "mtk_spacecraft.h"
#include  "mtk_label.h"
#include  "mtk_scale_line.h"
#include  "mtk_switches.h"
#include  "mtk_calc_alpha.h"




const float HITTEST_MIN_RADIUS = 5.0;	// ^[QbgI̍ŏsNZ



void
mtk_object::init(int type, int idx, float dz, bool bTargetable)
{
	m_type	= type;
	m_idx	= idx;
	m_dz	= dz;
	m_bTgt	= bTargetable; 
}


vector3d
mtk_object::get_pos() const
{
	return  get_object_positon( *this );
}


vector3d
mtk_object::get_parent_pos() const
{
	switch (m_type) {
		case OBJ_TYPE_SUN:
			return  8000.0 * vector3d(vGC);

		case OBJ_TYPE_SATELLITE:
			return  get_planet_position( get_satellite_planet_index(m_idx)  ); 

		case OBJ_TYPE_PLANET:
		default:
			return  vector3d(0,0,0);
	}
}



//------------------
//  Landing radius
//------------------
double
mtk_object::get_landing_radius()  const
{
	switch (m_type) {
		case OBJ_TYPE_SUN:				return  get_landing_radius_sun();
		case OBJ_TYPE_PLANET:			return  get_landing_radius_planet(m_idx);
		case OBJ_TYPE_SATELLITE:		return  get_landing_radius_satellite(m_idx);
		case OBJ_TYPE_SMALL_OBJECT:		return  get_landing_radius_small_object(m_idx);
		case OBJ_TYPE_SPACECRAFT:		return  get_landing_radius_spacecraft(m_idx);
		case OBJ_TYPE_STAR:				return  get_landing_radius_star(m_idx);
		case OBJ_TYPE_GLOBULAR_CLUSTER:	return  get_landing_radius_globular_cluster(m_idx);
		case OBJ_TYPE_OUR_GALAXY:		return  get_landing_radius_our_galaxy(m_idx);
		case OBJ_TYPE_NEARBY_GALAXY:	return  get_landing_radius_our_galaxy(m_idx);
		//case OBJ_TYPE_OORT_CLOUD:
		//case OBJ_TYPE_SCALE_CIRCLE:
		//case OBJ_TYPE_LABEL:
		default:
			return  0.0;
	}
}



//----------------
//  Prepare draw
//----------------
int
mtk_object::prepare_draw(const mtk_draw_struct &ds, const vector3d* nv, const float z_near, const float z_mid, const float z_far)
{
	int draw_rgn;
	float alp = 1.0;

	hit_test_status = OBJ_HITTEST_STAT_NONE;


	bool bFlg = true;
	switch (m_type) {
		case OBJ_TYPE_SUN:				bFlg = prepare_draw_sun(ds, &m_Rb, &m_Rh);							break;
		case OBJ_TYPE_PLANET:			bFlg = prepare_draw_planet(ds, m_idx, &m_Rb, &m_Rh);				break;
		case OBJ_TYPE_SATELLITE:		bFlg = prepare_draw_satellite(ds, m_idx, &m_Rb, &m_Rh);				break;
		case OBJ_TYPE_SMALL_OBJECT:		bFlg = prepare_draw_small_object(ds, m_idx, &m_Rb, &m_Rh);			break;
		case OBJ_TYPE_SPACECRAFT:		bFlg = prepare_draw_spacecraft(ds, m_idx, &m_Rb, &m_Rh);			break;
		case OBJ_TYPE_OORT_CLOUD:		bFlg = prepare_draw_Oort_cloud(ds, m_idx, &m_Rb, &m_Rh);			break;
		case OBJ_TYPE_STAR:				bFlg = prepare_draw_star(ds, m_idx, &m_Rb, &m_Rh);					break;
		case OBJ_TYPE_GLOBULAR_CLUSTER:	bFlg = prepare_draw_globular_cluster(ds, m_idx, &m_Rb, &m_Rh);		break;
		case OBJ_TYPE_OUR_GALAXY:		bFlg = prepare_draw_our_galaxy(ds, &m_Rb, &m_Rh);					break;
		case OBJ_TYPE_NEARBY_GALAXY:	bFlg = prepare_draw_nearby_galaxy(ds, m_idx, &m_Rb, &m_Rh, &alp);	break;
		case OBJ_TYPE_LABEL:			bFlg = prepare_draw_label(ds, m_idx, &m_Rb, &m_Rh, &alp);					break;

		case OBJ_TYPE_SCALE_CIRCLE:		bFlg = false;	break;	//*****
		default:						bFlg = false;	break;	//****
	}

	if (!bFlg)  return OBJ_DRAW_NONE;


	//****
	//--  IuWFNg̃J̑Έʒu
	const vector3d  pos = get_pos();
	const vector3d  cpos = ds.camera.get_pos();
	rpos = pos - cpos;
	const double l2 = rpos.norm2();

	//---  Draw Size  ---
	double  Rb = 1.1 * m_Rb;	// 傫߂Ɏ
	double  rpp = ds.rad_per_pixel;
	double  Dth2 = 4.0 * Rb*Rb/l2;
	double  psz2 = Dth2 / (rpp*rpp);


	double psz0 = (m_type == OBJ_TYPE_STAR)  ?  1.414  :  2.0 * min(ds.w, ds.h) / 600.0;
	double psz02 = psz0*psz0;




	alpha = 1.0f;
	if (m_type == OBJ_TYPE_STAR && psz2 < psz02/*2.0*/) {
		if ( tyStarMark == STARMARK_EXOPLANETS  && has_exoplanet( get_star(m_idx).get_HIP() ) ) {
			if (psz2 < 0.1)  return  OBJ_DRAW_NONE;
			alpha = psz2 / psz02;
			psz2 = psz02;
		} else {
			return  OBJ_DRAW_NONE;
		}
	}




	if (psz2 < psz0) {
		switch (m_type) {
			case OBJ_TYPE_STAR:
			case OBJ_TYPE_OORT_CLOUD:
				return OBJ_DRAW_NONE;
			default:
				break;
		}

		if (m_type == OBJ_TYPE_SUN) {
			draw_status = OBJ_DRAW_STAT_REALSIZE;
		} else {
			draw_status = OBJ_DRAW_STAT_POINT;

			const vector3d  ppos = get_parent_pos();
			double Dth, R, l;
			R = norm( pos - ppos );
			l = norm( ppos - cpos );
			Dth = 2.0 * R / l;
			alpha = calc_alpha_by_apparent_diameter( Dth );

			if (m_type == OBJ_TYPE_NEARBY_GALAXY) {
				alpha *= alp;
			}
			if (m_type == OBJ_TYPE_SMALL_OBJECT) {
				alpha *= get_minor_planet_point_alpha(ds, m_idx);
			}

			if (alpha < 0.1f)  return  OBJ_DRAW_NONE;
		}
	} else {
		draw_status = OBJ_DRAW_STAT_REALSIZE;
	}


	if (m_type == OBJ_TYPE_LABEL) {
		alpha = alp;
	}





	//-----------
	//  Culling
	//-----------
	vector3d rv = ds.calc_pos_on_viewframe( pos );
	double z = -rv.z;
	const double sf = ds.scale_factor;
	Rb *= sf;

	if (z + Rb < z_near)  return  OBJ_DRAW_NONE;
	for (int i=0; i<4; i++) {
		double d = dot_product(nv[i], rv);
				
		if (d > Rb)  return  OBJ_DRAW_NONE;
	}


	//--  z_far 艓̂͒uȂ
	double l = sf * sqrt(l2);
	scl = 1.0;
	const double l0 = z_far * 0.9;
	if (l > l0) {
		scl = l0 / l;
		rpos *= scl;
		z *= scl;
		Rb *= scl;
		l = l0;
	}


	//--  Region
	bool bNear, bFar;
	bNear = bFar = false;
	if (z - Rb < z_mid)	{ bNear = true; }
	if (z + Rb > z_mid)	{ bFar  = true; }

	if (bNear) {
		if (bFar)	draw_rgn = OBJ_DRAW_BOTH;
		else		draw_rgn = OBJ_DRAW_NEAR;
	} else if (bFar) {
		draw_rgn = OBJ_DRAW_FAR;
	} else {
		return  OBJ_DRAW_NONE;
	}


	//---  for z-list
	draw_z = z;


	//---  Hit-test  ---
	float  tgtc = alpha;

	if (m_type == OBJ_TYPE_STAR) {
		tgtc = 0.2 * psz2 / 2.0;
	}


	if ( !m_bTgt || tgtc < 0.5 ) {
		hit_test_status = OBJ_HITTEST_STAT_NONE;
		return  draw_rgn;
	}


	//---  Radius
	if (m_type != OBJ_TYPE_LABEL) {
		double  Rh = sf * m_Rh;
		double  Dth, psz;
		Dth = 2.0 * Rh / l;
		psz = Dth / ds.rad_per_pixel;
		hit_test_status  =  (psz > HITTEST_MIN_RADIUS)  ?  OBJ_HITTEST_STAT_REALSIZE  :  OBJ_HITTEST_STAT_POINT;
	} else {
		hit_test_status = OBJ_HITTEST_STAT_REALSIZE;
	}

	return  draw_rgn;
}



//---------
//  Scale
//---------
double
mtk_object::get_scale()  const
{
	switch (m_type) {
		case OBJ_TYPE_SUN:				return  get_scale_sun();
		case OBJ_TYPE_PLANET:			return  get_scale_planet(m_idx);
		case OBJ_TYPE_SATELLITE:		return  get_scale_satellite(m_idx);
		case OBJ_TYPE_SMALL_OBJECT:		return  get_scale_small_object(m_idx);
		case OBJ_TYPE_STAR:				return  get_scale_star(m_idx);
		case OBJ_TYPE_GLOBULAR_CLUSTER:	return  get_scale_globular_cluster(m_idx);
		case OBJ_TYPE_OUR_GALAXY:		return  get_scale_our_galaxy();
		case OBJ_TYPE_NEARBY_GALAXY:	return  get_scale_nearby_galaxy(m_idx);
		case OBJ_TYPE_SPACECRAFT:		return  get_scale_spacecraft(m_idx);
			break;
		default:
			break;
	}

	return  1000 * _1km_pc;	//****
}



//---------
//  Label
//---------
bool
is_object_label_visible(const mtk_draw_struct& ds, const mtk_object& obj)
{
	if (ds.tySubMode==SUBMODE_IMAGE || ds.tySubMode==SUBMODE_3DCHART)  return false;

	const int obj_type = obj.get_type();
	const int idx = obj.get_index();
	switch (obj_type) {
		case OBJ_TYPE_SUN:				return  is_sun_label_visible();
		case OBJ_TYPE_PLANET:			return  is_planet_label_visible(idx);
		case OBJ_TYPE_SATELLITE:		return  is_satellite_label_visible(idx);
		case OBJ_TYPE_SMALL_OBJECT:		return  is_small_object_label_visible(idx);
		case OBJ_TYPE_OORT_CLOUD:		return  is_Oort_cloud_label_visible();
		case OBJ_TYPE_STAR:				return  is_star_label_visible(ds, idx);
		case OBJ_TYPE_GLOBULAR_CLUSTER:	return  is_globular_cluster_label_visible(ds, idx);
		case OBJ_TYPE_NEARBY_GALAXY:	return  is_nearby_galaxy_label_visible(ds, idx);
		case OBJ_TYPE_SPACECRAFT:		return  is_spacecraft_label_visible(idx);

		case OBJ_TYPE_FIXED_OBJECT:		return  is_fixed_object_label_visible(ds, idx);
		case OBJ_TYPE_SCALE_CIRCLE:		return  is_scale_circle_label_visible( idx );

		default:						break;
	}

	return  false;
}


float
get_label_alpha(const mtk_draw_struct& ds, const mtk_object& obj)
{
	const int idx = obj.get_index();

	switch (obj.get_type()) {
		case OBJ_TYPE_SUN:				return  get_sun_label_alpha( ds );
		case OBJ_TYPE_PLANET:			return  get_planet_label_alpha( ds, idx );
		case OBJ_TYPE_SATELLITE:		return  get_satellite_label_alpha( ds, idx );
		case OBJ_TYPE_SMALL_OBJECT:		return  get_small_object_label_alpha( ds, idx );
		case OBJ_TYPE_OORT_CLOUD:		return  get_Oort_cloud_label_alpha( ds );
		case OBJ_TYPE_STAR:				return  get_star_label_alpha(ds, idx);
		case OBJ_TYPE_GLOBULAR_CLUSTER:	return  get_globular_cluster_label_alpha( ds, idx );
		case OBJ_TYPE_NEARBY_GALAXY:	return  get_nearby_galaxy_label_alpha(ds, idx);
		
		case OBJ_TYPE_SPACECRAFT:		return  get_spacecraft_label_alpha( ds, idx );
		case OBJ_TYPE_FIXED_OBJECT:		return	get_fixed_object_label_alpha( ds, idx );
		case OBJ_TYPE_SCALE_CIRCLE:		return  get_scale_circle_alpha( ds, idx );
		default:						break;
	}

	return  1.0f;
}




//--------
//  Draw
//--------
void
mtk_object::draw(const mtk_draw_struct& ds, bool bSelect, bool bChangeObjType)  const
{
	if (bChangeObjType && m_type != OBJ_TYPE_STAR) {
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
		glDisable(GL_LIGHTING);
		glDepthMask(GL_TRUE);
		glEnable(GL_DEPTH_TEST);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}


	if (draw_status == OBJ_DRAW_STAT_POINT) {
		switch (m_type) {
			case OBJ_TYPE_SMALL_OBJECT:		draw_a_small_object_point(ds, m_idx, rpos, alpha);		return;
			case OBJ_TYPE_NEARBY_GALAXY:	draw_a_nearby_galaxy_as_point(ds, m_idx, rpos, alpha);	return;
			default:	break;
		}


		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);

		if (m_type == OBJ_TYPE_GLOBULAR_CLUSTER || m_type == OBJ_TYPE_NEARBY_GALAXY) {
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		} else {
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}

		glPointSize(2);
		glColor3f(alpha, alpha, alpha);


		if (true) {
			// QUAD

			//---  JWn̎
			vector3d  eF, eU, eR;
			ds.camera.get_axes(eF, eU, eR);

			//---  xWn̎
			vector3d  eF2, eU2, eR2;
			double R2, R;
			R2 = norm2(rpos);
			if (R2 < 1e-100)  return;
			
			R = sqrt(R2);
			eF2 = rpos / R;
			eR2 = normalize( cross_product(eF2, eU) );
			eU2 = cross_product( eR2, eF2 );

			double sz;
			sz = R / get_screen_z() * get_screen_w() / 180;
			sz *= 0.4;

			vector3d r;
			glBegin(GL_QUADS);
				r = rpos + sz * eR2;
				glVertex3dv( r.get_pointer() );
				r = rpos + sz * eU2;
				glVertex3dv( r.get_pointer() );
				r = rpos - sz * eR2;
				glVertex3dv( r.get_pointer() );
				r = rpos - sz * eU2;
				glVertex3dv( r.get_pointer() );
			glEnd();
		} else {
			glBegin(GL_POINTS);
				glVertex3dv( rpos.get_pointer() );
			glEnd();
		}
	} else {
		switch (m_type) {
			case OBJ_TYPE_SUN:				draw_sun(ds, 0, rpos, scl);										break;
			case OBJ_TYPE_PLANET:			draw_a_planet(ds, m_idx, rpos, scl);							break;
			case OBJ_TYPE_SATELLITE:		draw_a_satellite(ds, m_idx, rpos, scl);							break;
			case OBJ_TYPE_SMALL_OBJECT:		draw_a_small_object(ds, m_idx, rpos, scl);						break;
			case OBJ_TYPE_SPACECRAFT:		draw_a_spacecraft(ds, m_idx, rpos, scl);						break;
			case OBJ_TYPE_OORT_CLOUD:		draw_Oort_cloud(ds, 0, rpos, scl);								break;
			case OBJ_TYPE_STAR:				draw_a_star(ds, m_idx, rpos, scl, m_Rb, alpha, bChangeObjType);	break;
			case OBJ_TYPE_GLOBULAR_CLUSTER:	draw_a_globular_cluster(ds, m_idx, rpos, scl);					break;
			case OBJ_TYPE_OUR_GALAXY:		draw_our_galaxy(ds, 0, rpos, scl);								break;
			case OBJ_TYPE_NEARBY_GALAXY:	draw_a_nearby_galaxy(ds, m_idx, rpos, scl);						break;

			case OBJ_TYPE_LABEL:			draw_a_label(ds, m_idx, rpos, scl, alpha, m_dz);				break;
			default:	break;
		}
	}


	return;
}




//------------
//  Hit-test
//------------
bool
mtk_object::hit_test(const mtk_draw_struct& ds, int mx, int my, const vector3d& cpos, const vector3d& nv) const
{
	//---------------
	//  x̏ꍇ
	//---------------
	if (m_type == OBJ_TYPE_LABEL) {
		return  hit_test_label(ds, m_idx, get_pos(), cpos, nv);
	}


	//----------------------------
	//  |CgTCYŔ肷ꍇ
	//----------------------------
	if (hit_test_status == OBJ_HITTEST_STAT_POINT) {
		const vector2f  pos = ds.calc_pos_on_screen( get_pos() );

		int ix, iy;
		ix = int(pos.x + 0.5);
		iy = int(pos.y + 0.5);

		int dx, dy;
		dx = ix - mx;
		dy = iy - my;
		if (dx*dx + dy*dy < HITTEST_MIN_RADIUS * HITTEST_MIN_RADIUS) {
			return  true;
		} else {
			return  false;
		}
	}


	//-----------------------
	//  TCYŔ肷ꍇ
	//-----------------------
	const vector3d pos  = get_pos();
	const vector3d rpos = cpos - pos;

	switch (m_type) {
		case  OBJ_TYPE_PLANET:			return  hit_test_planet(ds, m_idx, rpos, nv);
		case  OBJ_TYPE_SATELLITE:		return  hit_test_satellite(ds, m_idx, rpos, nv);
		case  OBJ_TYPE_SMALL_OBJECT:	return  hit_test_small_object(ds, m_idx, rpos, nv);
		case  OBJ_TYPE_OUR_GALAXY:		return  hit_test_our_galaxy(ds, rpos, nv);
		default:						break;
	}

	//--  IuWFNgƂ̌_߂
	const double b = dot_product(nv, rpos);
	const double D = b*b - rpos.norm2() + m_Rh*m_Rh;

	if (D <= 0.0)  return false;  // _Ȃ
	if (-b - sqrt(D) > 0.0)  return  true;

	return  false;
}














bool
is_object_orbit_visible(const mtk_object& obj)
{
	const int obj_type = obj.get_type();
	const int idx = obj.get_index();
	char dispsw_idx;


	switch (obj_type) {
		//--  Planets
		case OBJ_TYPE_PLANET:
			if (idx == 8)	dispsw_idx = DISPSW_TNO;
			else			dispsw_idx = DISPSW_PLANET;
			break;

		case OBJ_TYPE_SATELLITE:
			dispsw_idx = DISPSW_SATELLITE;
			break;
		
		//--  Small Objects
		case OBJ_TYPE_SMALL_OBJECT:
			{
				int type = get_small_object_type( idx );
				if (type == SMALLOBJ_TYPE_TNOS)		dispsw_idx = DISPSW_TNO;
				else								dispsw_idx = DISPSW_MINOR_PLANET;
			}
			break;


		case OBJ_TYPE_SCALE_CIRCLE:
			if (swScaleLine & SCALE_SQUARE && idx == 20) {
				return  true;
			}
			return  swScaleLine & SCALE_CIRCLE;

		default:
			return  true;
	}


	if ( !test_disp_switch( dispsw_idx, DISPM_DISP) || !test_disp_switch( dispsw_idx, DISPM_ORBIT ) )  return  false;


	if (obj_type == OBJ_TYPE_SATELLITE) {
		if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_satellite_selected(idx)) {
			return  false;
		}
	}

	if (obj_type == OBJ_TYPE_SMALL_OBJECT) {
		if (test_disp_switch(dispsw_idx, DISPM_SELECT) && !is_small_object_selected(idx)) {
			return  false;
		}
	}


	return  true;
}




//---------
//  Alpha
//---------

float
get_orbit_alpha(const mtk_draw_struct& ds, const mtk_object& obj)
{
	int idx = obj.get_index();

	switch (obj.get_type()) {
		case OBJ_TYPE_PLANET:		return  get_planet_orbit_alpha( ds, idx );
		case OBJ_TYPE_SATELLITE:	return  get_satellite_orbit_alpha( ds, idx );
		case OBJ_TYPE_SMALL_OBJECT:	return  get_small_object_orbit_alpha( ds, idx );
		case OBJ_TYPE_SCALE_CIRCLE:	return  get_scale_circle_alpha( ds, idx );
		default:				break;
	}

	return  1.0f;
}






//--------
//  Info
//--------
void
mtk_object::make_object_info(const mtk_draw_struct& ds, std::vector< std::string >& text)  const
{
	text.clear();

	switch (m_type) {
		case OBJ_TYPE_SUN:				make_object_info_sun(ds, text);							break;
		case OBJ_TYPE_PLANET:			make_object_info_planet(ds, m_idx, text);				break;
		case OBJ_TYPE_SATELLITE:		make_object_info_satellite(ds, m_idx, text);			break;
		case OBJ_TYPE_SMALL_OBJECT:		make_object_info_small_object(ds, m_idx, text);			break;
		case OBJ_TYPE_STAR:				make_object_info_star(ds, m_idx, text);					break;
		case OBJ_TYPE_GLOBULAR_CLUSTER:	make_object_info_globular_cluster(ds, m_idx, text);		break;
		case OBJ_TYPE_OUR_GALAXY:		make_object_info_our_galaxy(ds, text);					break;
		case OBJ_TYPE_NEARBY_GALAXY:	make_object_info_nearby_galaxy(ds, m_idx, text);		break;
		case OBJ_TYPE_SPACECRAFT:		make_object_info_spacecraft(ds, m_idx, text);			break;

		default:
			if (ds.lang == LANG_ENG) {
				text.push_back( "None" );
				text.push_back( "None" );
			} else {
				text.push_back( "`" );
				text.push_back( "`" );
			}
			break;
	}
}
