/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_planet_orientation.h"
#include  "../common/astro/astrometry.h"
#include  "mtk_planet_position.h"
#include  "mtk_satellite_data.h"





//-------------
//  Constants  ***
//-------------
const int PlanetNum = 9;

//---------
//  Array
//---------
static matrix3d  vPlanetEcToPlMatrix0[PlanetNum];	// without planet rotation
static matrix3d  vPlanetEcToPlMatrix[PlanetNum];	// with planet rotaiton


static matrix3d  vSatelliteEcToSatMatrix[300];		//*****



//----------
//  update
//----------
void	update_planets_orientaion(const double& t);
void	update_satellites_orientaion(const double& t);

void
update_planets_and_satellites_orientation(const double& t)
{
	const double DT_MIN = 0.01/(24.0*60.0*60.0);
	static bool first = true;
	static double t_last;

	if (first)	{ first = false; }
	else { if (fabs(t - t_last) < DT_MIN)  return; }

	update_planets_orientaion(t);
	update_satellites_orientaion(t);

	t_last = t;
}


//===========
//  Planets
//===========

//----------------------------------------
//  f̎]߂iԌoEԈ܁j Ref[1]
//----------------------------------------
void
calc_planet_orientation_angle(int idx, const double& t, double& alp, double& dlt, double& W)
{
	const double  JD0 = 2451545.0;	//  2000/1/1.5
	const double  d = t - JD0;
	const double  T = d/36525.0;

	switch (idx) {
		case 0:	// 
			alp = 281.01 - 0.033*T;
			dlt = 61.45  - 0.005*T;
			W   = 329.548 + 6.1385025*d;
			break;
		case 1:	// 
			alp = 272.76;
			dlt = 67.16;
			W   = 160.20 - 1.4813688*d;
			break;
		case 2:	// n
			alp = 0.00 - 0.641*T;
			dlt = 90.00 - 0.557*T;
			W   = 190.147 + 360.9856235*d;
			break;
		case 3:	// ΐ
			alp = 317.68143 - 0.1061*T;
			dlt = 52.88650  - 0.0609*T;
			W   = 176.630   + 350.89198226*d;
			break;
		case 4:	// ؐ
			alp = 268.05 - 0.009*T;
			dlt = 64.49  + 0.003*T;
			W   = 284.95 + 870.5366420*d;
			break;
		case 5:	// y
			alp = 40.589 - 0.036*T;
			dlt = 83.537 - 0.004*T;
			W   = 38.90  + 810.7939024*d;
			break;
		case 6:	// V
			alp = 257.311;
			dlt = -15.175;
			W   = 203.81 - 501.1600928*d;
			break;
		case 7:	// C
			{
				double N, th, sinN;
				N = 357.85 + 52.316*T;
				th = N*D2R;
				sinN = sin(th);
				alp = 299.36 + 0.70*sinN;
				dlt = 43.46 - 0.51*cos(th);
				W   = 253.18 + 536.3128492*d - 0.48*sinN;
			}
			break;
		case 8:	// 
			alp = 313.02;
			dlt = 9.09;
			W   = 236.77 - 56.3623195*d;
			break;
		default:
			alp = dlt = W = 0;
			break;
	}
}


//------------------------------------------
//  WnfWn֕ϊs𓾂
//------------------------------------------
void
calc_planet_ec_to_pl_matrix(int idx, const double& t, matrix3d& M0, matrix3d& M)
{
	static matrix3d  M_J2000;
	static bool first = true;
	if (first) {
		first = false;
		M_J2000 = Rx3deg(epsJ2000);
	}

	double alp, dlt, W;
	calc_planet_orientation_angle(idx, t, alp, dlt, W);

	M0 = Rzxz3deg(0.0, -90+dlt, -90-alp) * M_J2000;
	M  = Rzxz3deg( -W, -90+dlt, -90-alp) * M_J2000;
}

void
update_planets_orientaion(const double& t)
{
	for (int i=0; i<PlanetNum; i++) {
		matrix3d  M0, M;
		calc_planet_ec_to_pl_matrix(i, t, M0, M);
		vPlanetEcToPlMatrix0[i] = M0;
		vPlanetEcToPlMatrix[i]  = M;
	}
}

matrix3d
get_planet_ec_to_pl_matrix0(int idx)
{
	return  vPlanetEcToPlMatrix0[idx];
}

matrix3d
get_planet_pl_to_ec_matrix0(int idx)
{
	return  vPlanetEcToPlMatrix0[idx].transposed();
}

matrix3d
get_planet_ec_to_pl_matrix(int idx)
{
	return  vPlanetEcToPlMatrix[idx];
}

matrix3d
get_planet_pl_to_ec_matrix(int idx)
{
	return  vPlanetEcToPlMatrix[idx].transposed();
}



//------------------------------------------------
//  WnɂAfWn̊xNg߂
//------------------------------------------------
//   ez: kɕ
//   ex: ԓEqߐV
//   ey: ԓEqߐ㓌
//------------------------------------------------
void
get_planet_bases(int idx, vector3d& ex, vector3d& ey, vector3d& ez)
{
	const matrix3d  A = vPlanetEcToPlMatrix[idx];
	ex.set(A.m11, A.m12, A.m13);
	ey.set(A.m21, A.m22, A.m23);
	ez.set(A.m31, A.m32, A.m33);
}


//----------------------------------------
//  Wn̈ʒuxNgfWnɕϊ
//----------------------------------------
vector3d
ec_to_pl(int idx, const vector3d&  pos_ec)
{
	vector3d  pos_pl = pos_ec - get_planet_position(idx);	// fSł̈ʒu

	return vPlanetEcToPlMatrix[idx] * pos_pl;
}


//------------------------------------------------
//  fSxNg nv ̌̕oxEܓx߂
//  nv ͉̐W(J2000)ł̒lƂ
//------------------------------------------------
vector2d
get_planet_lon_lat_ec(int idx, const vector3d& nv)
{
	//--  fWñxNg
	vector3d  ex, ey, ez;
	get_planet_bases(idx, ex, ey, ez);

	//--  lon, lat ߂
	double x, y, z, rho;
	double th, phi;
	double lon, lat;
	x = dot_product(nv, ex);
	y = dot_product(nv, ey);
	z = dot_product(nv, ez);
	rho = sqrt(x*x + y*y);

	th  = atan2(rho, z);
	phi = atan2(y, x);

	lon = R2D * phi;
	lat = 90.0 - R2D * th;
	if (lon > 180.0) lon -= 360.0;

	return  vector2d(lon, lat);
}


//---------------------------------------------------------
//  ܓxEoxAfWnł̉̒PʃxNg߂
//---------------------------------------------------------
vector3d
get_planet_nv_pl(const double& lon, const double& lat)
{
	const double th  = D2R * (90.0 - lat);
	const double phi = D2R * lon;
	const double rho = sin(th);

	return  vector3d(rho*cos(phi), rho*sin(phi), cos(th));
}

//---------------------------------------------------------------
//  oxEܓxA̒PʃxNg߂iWn(J2000)j
//---------------------------------------------------------------
vector3d
get_planet_nv_ec(int idx, const double& lon, const double& lat)
{
	return  get_planet_ec_to_pl_matrix(idx).transposed() * get_planet_nv_pl(lon, lat);
}


//---------------------------------------------------------
//  vl^E[h̉Wnł̊Wn𓾂
//---------------------------------------------------------
quaterniond
get_planetarium_ori(int idx, const double& lon, const double& lat, const double& azim, const double& alti)
{
	vector3d  nv  = get_planet_nv_ec(idx, lon, lat);
	vector3d  nnv = get_planet_nv_ec(idx, 0,   90.0);
	vector3d  ex = normalize( cross_product(nnv, nv) );	// 
	vector3d  ey = normalize( cross_product(ex,  nv) );	// 

	// ʊp
	const double  phi = D2R * azim;
	vector3d _ex, _ey;
	_ex = cos(phi) * ex - sin(phi) * ey;
	_ey = cos(phi) * ey + sin(phi) * ex;

	// x
	const double  th = D2R * alti;
	vector3d eF, eU;
	eF = cos(th) * _ey + sin(th) *  nv;
	eU = cos(th) *  nv - sin(th) * _ey;

	return  axes_to_quaternion(eF, eU);
}

//---------------------------------------------------------
//  vl^E[hł̐ʕAWnł̎p𓾂
//---------------------------------------------------------
quaterniond
get_ori_from_planetarium_dirc(int idx, const double& lon, const double& lat, const double& thd)
{
	vector3d  nv  = get_planet_nv_ec(idx, lon, lat);
	vector3d  nnv = get_planet_nv_ec(idx, 0, 90.0);
	vector3d  ex = normalize( cross_product(nnv, nv) );	// 
	vector3d  ey = normalize( cross_product(ex,  nv) );	// 

	const double  th = D2R * thd;
	vector3d  eF = cos(th) * ey + sin(th) * ex;
	quaterniond  ori = axes_to_quaternion(-nv, eF);

	return  ori;
}


//---------------------------------------------------------
//  RD[hł̎pAvl^EE[h̐ʕ𓾂
//---------------------------------------------------------
double
get_dirc_from_3D_ori(int idx, const quaterniond& ori)
{
	vector3d  eF, eU, eR;
	quaternion_to_axes(ori, eF, eU, eR);

	vector3d  nnv = get_planet_nv_ec(idx, 0, 90.0);
	double x, y, th;
	x = dot_product(-eR, nnv);
	y = dot_product(-eU, nnv);
	th = atan2(y, x);

	return  90.0 - R2D * th;
}



//==============
//  Satellites
//==============

//-------------------------
//  ̌߂  Ref[1]
//-------------------------
void
calc_moon_orientation_angle(const double& t, double& alp, double& dlt, double& W)
{
	const double  JD0 = 2451545.0;	//  2000/1/1.5
	const double  d = t - JD0;
	const double  T = d/36525.0;

	double E1, E2, E3, E4, E5, E6, E7, E8, E9, E10, E11, E12, E13;
	E1  = 125.045 - 0.0529921 *d;
	E2  = 250.089 - 0.1059842 *d;
	E3  = 260.008 + 13.0120009*d;
	E4  = 176.625 + 13.3407154*d;
	E5  = 357.529 + 0.9856003 *d;
	E6  = 311.589 + 26.4057084*d;
	E7  = 134.963 + 13.0649930*d;
	E8  = 276.617 + 0.3287146 *d;
	E9  = 34.226  + 1.7484877 *d;
	E10 = 15.134  - 0.1589763 *d;
	E11 = 119.743 + 0.0036096 *d;
	E12 = 239.961 + 0.1643573 *d;
	E13 = 25.053  + 12.9590088*d;

	const double& a = D2R;
	double sinE1, sinE2, sinE3, sinE4, sinE5, sinE6, sinE7;
	double sinE8, sinE9, sinE10, sinE11, sinE12, sinE13;
	double cosE1, cosE2, cosE3, cosE4,  cosE6, cosE7,  cosE10,  cosE13;
	sinE1  = sin(a*E1);
	sinE2  = sin(a*E2);
	sinE3  = sin(a*E3);
	sinE4  = sin(a*E4);
	sinE5  = sin(a*E5);
	sinE6  = sin(a*E6);
	sinE7  = sin(a*E7);
	sinE8  = sin(a*E8);
	sinE9  = sin(a*E9);
	sinE10 = sin(a*E10);
	sinE11 = sin(a*E11);
	sinE12 = sin(a*E12);
	sinE13 = sin(a*E13);
	cosE1  = cos(a*E1);
	cosE2  = cos(a*E2);
	cosE3  = cos(a*E3);
	cosE4  = cos(a*E4);
	cosE6  = cos(a*E6);
	cosE7  = cos(a*E7);
	cosE10 = cos(a*E10);
	cosE13 = cos(a*E13);

	alp = 269.9949	+ 0.0031*T		- 3.8787*sinE1	- 0.1204*sinE2
					+ 0.0700*sinE3	- 0.0172*sinE4	+ 0.0072*sinE6
					- 0.0052*sinE10	+ 0.0043*sinE13;
	dlt	= 66.5392	+ 0.0130*T		+ 1.5419*cosE1	+ 0.0239*cosE2
					- 0.0278*cosE3	+ 0.0068*cosE4	- 0.0029*cosE6
					+ 0.0009*cosE7	+ 0.0008*cosE10	- 0.0009*cosE13;
	W	= 38.3213	+ 13.17635815*d	- 1.4e-12 *d*d	+ 3.5610*sinE1
					+ 0.1208*sinE2	- 0.0642*sinE3	+ 0.0158*sinE4
					+ 0.0252*sinE5	- 0.0066*sinE6	- 0.0047*sinE7
					- 0.0046*sinE8	+ 0.0028*sinE9	+ 0.0052*sinE10
					+ 0.0040*sinE11	+ 0.0019*sinE12	- 0.0044*sinE13;
}

//---------------------------
//  q̌߂  Ref[1]
//---------------------------
void
calc_satellite_orientation_angle(int idx, const double& t, double& alp, double& dlt, double& W)
{
	if (idx == SAT_MOON) {
		calc_moon_orientation_angle(t, alp, dlt, W);
		return;
	}

	const double  JD0 = 2451545.0;	//  2000/1/1.5
	const double  d = t - JD0;
	const double  T = d/36525.0;
	const double  T2 = T*T;

	switch (idx) {
	//== Mars ==
		case SAT_PHOBOS:	// Phobos
			{
				const double M1 = D2R * (169.51 - 0.4357640*d);
				const double M2 = D2R * (192.93 + 1128.4096700*d + 8.864*T2);
				const double sinM1 = sin(M1);
				alp = 317.68 - 0.108*T + 1.79*sinM1;
				dlt =  52.90 - 0.061*T - 1.08*cos(M1);
				W   = 35.06  + 1128.8445850*d + 8.864*T2 - 1.42*sinM1 - 0.78*sin(M2);
			}
			break;
		case SAT_DEIMOS:	// Deimos
			{
				const double M3 = D2R * (53.47 - 0.0181510*d);
				const double sinM3 = sin(M3);
				const double cosM3 = cos(M3);
				alp = 316.65 - 0.108*T + 2.98*sinM3;
				dlt =  53.52 - 0.061*T - 1.78*cosM3;
				W   =  79.41 + 285.1618970*d - 0.520*T2 - 2.58*sinM3 + 0.19*cosM3;
			}
			break;
	//== Jupiter ==
		case SAT_IO:	// Io
			{
				const double J3 = D2R * (283.90 + 4850.7*T);
				const double J4 = D2R * (355.80 + 1191.3*T);
				const double sinJ3 = sin(J3);
				const double sinJ4 = sin(J4);
				alp = 268.05 - 0.009*T + 0.094*sinJ3   + 0.024*sinJ4;
				dlt =  64.50 + 0.003*T + 0.040*cos(J3) + 0.011*cos(J4);
				W   = 200.39 + 203.4889538*d - 0.085*sinJ3 - 0.022*sinJ4;
			}
            break;
		case SAT_EUROPA:	// Europa
			{
				const double J4 = D2R * (355.80 + 1191.3*T);
				const double J5 = D2R * (119.90 +  262.1*T);
				const double J6 = D2R * (229.80 +   64.3*T);
				const double J7 = D2R * (352.35 + 2382.6*T);
				const double sinJ4 = sin(J4);
				const double sinJ5 = sin(J5);
				const double sinJ6 = sin(J6);
				const double sinJ7 = sin(J7);
				alp = 268.08 - 0.009*T + 1.086*sinJ4   + 0.060*sinJ5   + 0.015*sinJ6   + 0.009*sinJ7;
				dlt =  64.51 + 0.003*T + 0.468*cos(J4) + 0.026*cos(J5) + 0.007*cos(J6) + 0.002*cos(J7);
				W   =  36.022 + 101.3747235*d - 0.980*sinJ4 - 0.054*sinJ5 - 0.014*sinJ6 - 0.008*sinJ7;
			}
            break;
		case SAT_GANYMEDE:	// Ganymede
			{
				const double J4 = D2R * (355.80 + 1191.3*T);
				const double J5 = D2R * (119.90 +  262.1*T);
				const double J6 = D2R * (229.80 +   64.3*T);
				const double sinJ4 = sin(J4);
				const double sinJ5 = sin(J5);
				const double sinJ6 = sin(J6);
				alp = 268.20 - 0.009*T - 0.037*sinJ4   + 0.431*sinJ5   + 0.091*sinJ6;
				dlt =  64.57 + 0.003*T - 0.016*cos(J4) + 0.186*cos(J5) + 0.039*cos(J6);
				W   =  44.064 + 50.3176081*d - 0.033*sinJ4 - 0.389*sinJ5 - 0.082*sinJ6;
			}
            break;
		case SAT_CALLISTO:	// Callisto
			{
				const double J5 = D2R * (119.90 +  262.1*T);
				const double J6 = D2R * (229.80 +   64.3*T);
				const double J8 = D2R * (113.35 + 6070.0*T);
				const double sinJ5 = sin(J5);
				const double sinJ6 = sin(J6);
				const double sinJ8 = sin(J8);
				alp = 268.72 - 0.009*T - 0.068*sinJ5   + 0.590*sinJ6   + 0.010*sinJ8;
				dlt =  64.83 + 0.003*T - 0.029*cos(J5) + 0.254*cos(J6) - 0.004*cos(J8);
				W   = 259.51 + 21.5710715*d + 0.061*sinJ5 - 0.533*sinJ6 - 0.009*sinJ8;
			}
            break;
	//== Pluto ==
		case SAT_CHARON:	// Charon
			alp = 313.02;
			dlt = 9.09;
			W   = 56.77 - 56.3623195*d;
			break;

		default:
			alp = W = 0.0;
			dlt = 90.0;
			break;
	}
}



matrix3d
calc_satellite_ec_to_sat_matrix(int idx, const double& t)
{
	static matrix3d  M_J2000;
	static bool first = true;
	if (first) {
		first = false;
		M_J2000 = Rx3deg(epsJ2000);
	}

	// q̉]
	double alp, dlt, W;
	calc_satellite_orientation_angle(idx, t, alp, dlt, W);

	return  Rzxz3deg(-W,-90+dlt,-90-alp) * M_J2000;
}


void
update_satellites_orientaion(const double& t)
{
	for (int i=0; i<get_satellites_num(); i++) {
		vSatelliteEcToSatMatrix[i] = calc_satellite_ec_to_sat_matrix(i, t);
	}
}


matrix3d
get_satellite_ec_to_sat_matrix(int idx)
{
	return  vSatelliteEcToSatMatrix[idx];
}

matrix3d
get_satellite_sat_to_ec_matrix(int idx)
{
	return  vSatelliteEcToSatMatrix[idx].transposed();
}

//------------------------------------------------
//  WnɂAqWn̊xNg߂
//------------------------------------------------
//   ez: kɕ
//   ex: ԓEqߐV
//   ey: ԓEqߐ㓌
//------------------------------------------------
void
get_satellite_bases(int idx, vector3d& ex, vector3d& ey, vector3d& ez)
{
	const matrix3d  A = vSatelliteEcToSatMatrix[idx];
	ex.set(A.m11, A.m12, A.m13);
	ey.set(A.m21, A.m22, A.m23);
	ez.set(A.m31, A.m32, A.m33);
}




////////////////////////////////////////
/// References
//
// [1] Seidelmann et al., 2002, Celestial Mechanics and Dynamical Astronomy, 82, 83-100
